/*
 *  arch/s390/kernel/time.c
 *
 *  S390 version
 *    Copyright (C) 1999 IBM Deutschland Entwicklung GmbH, IBM Corporation
 *    Author(s): Hartmut Penner (hp@de.ibm.com),
 *               Martin Schwidefsky (schwidefsky@de.ibm.com),
 *               Denis Joseph Barrow (djbarrow@de.ibm.com,barrow_dj@yahoo.com)
 *
 *  Derived from "arch/i386/kernel/time.c"
 *    Copyright (C) 1991, 1992, 1995  Linus Torvalds
 */

#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/param.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/interrupt.h>
#include <linux/time.h>
#include <linux/timer.h>
#include <linux/delay.h>
#include <linux/init.h>
#include <linux/smp.h>
#include <linux/types.h>

#include <asm/uaccess.h>
#include <asm/delay.h>
#include <asm/s390_ext.h>

#include <linux/timex.h>
#include <linux/config.h>

#include <asm/irq.h>


/* change this if you have some constant time drift */
#define USECS_PER_JIFFY     ((unsigned long) 1000000/HZ)
#define CLK_TICKS_PER_JIFFY ((unsigned long) USECS_PER_JIFFY << 12)

#define TICK_SIZE tick

uint64_t init_timer_cc;
static ext_int_info_t ext_int_info_comparator; 
#ifdef CONFIG_NO_HZ_TIMER
static ext_int_info_t ext_int_info_timer; 
#endif

extern rwlock_t xtime_lock;
extern unsigned long wall_jiffies;

void tod_to_timeval(__u64 todval, struct timeval *xtime)
{
        const int high_bit = 0x80000000L;
        const int c_f4240 = 0xf4240L;
        const int c_7a120 = 0x7a120;
	/* We have to divide the 64 bit value todval by 4096
	 * (because the 2^12 bit is the one that changes every 
         * microsecond) and then split it into seconds and
         * microseconds. A value of max (2^52-1) divided by
         * the value 0xF4240 can yield a max result of approx
         * (2^32.068). Thats to big to fit into a signed int
	 *   ... hacking time!
         */
	asm volatile ("L     2,%1\n\t"
		      "LR    3,2\n\t"
		      "SRL   2,12\n\t"
		      "SLL   3,20\n\t"
		      "L     4,%O1+4(%R1)\n\t"
		      "SRL   4,12\n\t"
		      "OR    3,4\n\t"  /* now R2/R3 contain (todval >> 12) */
		      "SR    4,4\n\t"
		      "CL    2,%2\n\t"
		      "JL    .+12\n\t"
		      "S     2,%2\n\t"
		      "L     4,%3\n\t"
                      "D     2,%4\n\t"
		      "OR    3,4\n\t"
		      "ST    2,%O0+4(%R0)\n\t"
		      "ST    3,%0"
		      : "=m" (*xtime) : "m" (todval),
		        "m" (c_7a120), "m" (high_bit), "m" (c_f4240)
		      : "cc", "memory", "2", "3", "4" );
}

static inline unsigned long do_gettimeoffset(void) 
{
	__u64 now;

	asm ("STCK 0(%0)" : : "a" (&now) : "memory");
        now = (now - init_timer_cc) >> 12;
	/* We require the offset from the latest update of xtime */
	now -= (__u64) wall_jiffies*USECS_PER_JIFFY;
	return (unsigned long) now;
}

/*
 * This version of gettimeofday has microsecond resolution.
 */
void do_gettimeofday(struct timeval *tv)
{
	unsigned long flags;
	unsigned long usec, sec;

	read_lock_irqsave(&xtime_lock, flags);
	sec = xtime.tv_sec;
	usec = xtime.tv_usec + do_gettimeoffset();
	read_unlock_irqrestore(&xtime_lock, flags);

	while (usec >= 1000000) {
		usec -= 1000000;
		sec++;
	}

	tv->tv_sec = sec;
	tv->tv_usec = usec;
}

void do_settimeofday(struct timeval *tv)
{

	write_lock_irq(&xtime_lock);
	/* This is revolting. We need to set the xtime.tv_usec
	 * correctly. However, the value in this location is
	 * is value at the last tick.
	 * Discover what correction gettimeofday
	 * would have done, and then undo it!
	 */
	tv->tv_usec -= do_gettimeoffset();

	while (tv->tv_usec < 0) {
		tv->tv_usec += 1000000;
		tv->tv_sec--;
	}

	xtime = *tv;
	time_adjust = 0;		/* stop active adjtime() */
	time_status |= STA_UNSYNC;
	time_maxerror = NTP_PHASE_LIMIT;
	time_esterror = NTP_PHASE_LIMIT;
	write_unlock_irq(&xtime_lock);
}

#ifdef CONFIG_NO_HZ_TIMER
/*
 * On S/390 we don't have the regular 1/100 HZ tick anymore.
 * The clock comparator will interrupt at the next timer event.
 * time_warp is called whenever a processor enters the kernel and
 * the tod clock contains value greater then jiffy_timer (see entry.S).
 */

void time_warp(void) {
	struct task_struct *p = current;
	int cpu = smp_processor_id();
	int user, sys;
        __u64 tmp;

	tmp = S390_lowcore.last_mark >> 12;
	tmp = do_div(tmp, USECS_PER_JIFFY) << 12;
	S390_lowcore.jiffy_timer = S390_lowcore.last_mark
		- tmp + CLK_TICKS_PER_JIFFY;
	/* Update xtime structure */
	update_times_irqsave();
	/* Update user process tick count */
	tmp = p->thread.user_time >> 12;
	do_div(tmp, USECS_PER_JIFFY);
	user = tmp - p->times.tms_utime;
	tmp = p->thread.system_time >> 12;
        do_div(tmp, USECS_PER_JIFFY);
        sys = tmp - p->times.tms_stime;
	/*
	 * update_process_times() expects us to have done irq_enter().
	 * Besides, if we don't timer interrupts ignore the global
	 * interrupt lock, which is the WrongThing (tm) to do.
	 */
	irq_enter(cpu, 0);
	update_process_times2(user, sys);
	irq_exit(cpu, 0);
}

atomic_t tq_timer_active = ATOMIC_INIT(0);
struct timer_list tq_timer_list;

static void tq_timer_event(unsigned long __data)
{
	del_timer(&tq_timer_list);
	atomic_set(&tq_timer_active, 0);
	mark_bh(TQUEUE_BH);
}

void tq_pending(void) {
	if (atomic_compare_and_swap(0, 1, &tq_timer_active) != 0)
		return;
	/* setup timer event for the next HZ tick.  */
	init_timer(&tq_timer_list);
	tq_timer_list.expires = jiffies + 1;
	tq_timer_list.function = tq_timer_event;
	add_timer(&tq_timer_list);
}

static void do_comparator_interrupt(struct pt_regs *regs, __u16 error_code)
{
	int cpu = smp_processor_id();

	irq_enter(cpu, 0);
	S390_lowcore.timer_event = (__u64) -1;
	asm volatile ("SCKC %0" : : "m" (S390_lowcore.timer_event));
	mark_bh(TIMER_BH);
	irq_exit(cpu, 0);
}

static void s390_timer_notify(unsigned long expires)
{
        S390_lowcore.timer_event = 
		((__u64) expires*CLK_TICKS_PER_JIFFY) + init_timer_cc;
	asm volatile ("SCKC %0" : : "m" (S390_lowcore.timer_event));
}

/*
 * Linux uses the S/390 cpu timer interrupt (one shot interrupt)
 * to get an interrupt when the time slice of the current process
 * ends. switch_to() calls set_time_slice to update the cpu
 * timer register. In fact system_enter should already have called
 * update_process_times2 so there is nothing to do. Just to make
 * sure we set need_resched.
 */
void set_cpu_timer(void)
{
	unsigned long min_ticks;
	__u64 time_slice;
	if (current->pid != 0 && current->need_resched == 0) {
		min_ticks = current->counter;
		if (current->it_prof_value > 0 &&
		    current->it_prof_value < min_ticks)
			min_ticks = current->it_prof_value;
		if (current->it_virt_value > 0 &&
		    current->it_virt_value < min_ticks)
			min_ticks = current->it_virt_value;
		time_slice = (__u64) min_ticks*CLK_TICKS_PER_JIFFY;
	} else
		time_slice = 0x7ffffffffffff000LL;
	asm volatile ("spt %0" : : "m" (time_slice));
} 

void do_timer_interrupt(struct pt_regs *regs, __u16 error_code)
{
	set_cpu_timer();
}

void init_cpu_timer(void)
{
	unsigned long cr0;
	__u64 big_spt = 0x7ffffffffffff000LL;

	/* set cpu timer to something big */
	asm volatile ("spt %0" : : "m" (big_spt));
	S390_lowcore.timer_event = (__u64) -1;
	asm volatile ("SCKC %0" : : "m" (S390_lowcore.timer_event));
        S390_lowcore.jiffy_timer = init_timer_cc + CLK_TICKS_PER_JIFFY;
        S390_lowcore.last_mark = init_timer_cc;
        /* allow clock comparator timer interrupt */
        asm volatile ("STCTL 0,0,%0" : "=m" (cr0) : : "memory");
        cr0 |= 0xc00;
        asm volatile ("LCTL 0,0,%0" : : "m" (cr0) : "memory");
}
#else
/*
 * timer_interrupt() needs to keep up the real-time clock,
 * as well as call the "do_timer()" routine every clocktick
 */

#ifdef CONFIG_SMP
extern __u16 boot_cpu_addr;
#endif

static void do_comparator_interrupt(struct pt_regs *regs, __u16 error_code)
{
	int cpu = smp_processor_id();

	irq_enter(cpu, 0);

	/*
	 * set clock comparator for next tick
	 */
        S390_lowcore.jiffy_timer += CLK_TICKS_PER_JIFFY;
        asm volatile ("SCKC %0" : : "m" (S390_lowcore.jiffy_timer));

#ifdef CONFIG_SMP
	if (S390_lowcore.cpu_data.cpu_addr == boot_cpu_addr)
		write_lock(&xtime_lock);

	update_process_times(user_mode(regs));

	if (S390_lowcore.cpu_data.cpu_addr == boot_cpu_addr) {
		do_timer(regs);
		write_unlock(&xtime_lock);
	}
#else
	do_timer(regs);
#endif

	irq_exit(cpu, 0);
}

void init_cpu_timer(void)
{
	unsigned long cr0;

	S390_lowcore.jiffy_timer = (__u64) jiffies * CLK_TICKS_PER_JIFFY;
	S390_lowcore.jiffy_timer += init_timer_cc + CLK_TICKS_PER_JIFFY;
	asm volatile ("SCKC %0" : : "m" (S390_lowcore.jiffy_timer));
        /* allow clock comparator timer interrupt */
        asm volatile ("STCTL 0,0,%0" : "=m" (cr0) : : "memory");
        cr0 |= 0x800;
        asm volatile ("LCTL 0,0,%0" : : "m" (cr0) : "memory");
}
#endif

/*
 * Initialize the TOD clock and the CPU timer of
 * the boot cpu.
 */
void __init time_init(void)
{
        __u64 set_time_cc;
	int cc;

        /* kick the TOD clock */
        asm volatile ("STCK 0(%1)\n\t"
                      "IPM  %0\n\t"
                      "SRL  %0,28" : "=r" (cc) : "a" (&init_timer_cc) 
				   : "memory", "cc");
        switch (cc) {
        case 0: /* clock in set state: all is fine */
                break;
        case 1: /* clock in non-set state: FIXME */
                printk("time_init: TOD clock in non-set state\n");
                break;
        case 2: /* clock in error state: FIXME */
                printk("time_init: TOD clock in error state\n");
                break;
        case 3: /* clock in stopped or not-operational state: FIXME */
                printk("time_init: TOD clock stopped/non-operational\n");
                break;
        }

	/* set xtime */
        set_time_cc = init_timer_cc - 0x8126d60e46000000LL +
                      (0x3c26700LL*1000000*4096);
        tod_to_timeval(set_time_cc, &xtime);

        /* request the 0x1004 external interrupt */
        if (register_early_external_interrupt(0x1004, do_comparator_interrupt,
                                             &ext_int_info_comparator) != 0)
                panic("Couldn't request external interrupt 0x1004");

#ifdef CONFIG_NO_HZ_TIMER
        /* request the 0x1005 external interrupt */
        if (register_early_external_interrupt(0x1005, do_timer_interrupt,
					      &ext_int_info_timer) != 0)
                panic("Couldn't request external interrupt 0x1005");

	/* set the notifier callback */
	timer_notify = s390_timer_notify;
        itimer_notify = set_cpu_timer;
#endif

        /* init CPU timer */
        init_cpu_timer();

	/* Set do_get_fast_time function pointer.  */
	do_get_fast_time = do_gettimeofday;
}
