/*
 * hangcheck-timer.c
 *
 * Test driver for a little io fencing timer.
 *
 * Copyright (C) 2002 Oracle Corporation.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have recieved a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 021110-1307, USA.
 */


#include <linux/module.h>
#include <linux/config.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/reboot.h>
#include <linux/smp_lock.h>
#include <linux/init.h>
#include <asm/uaccess.h>

#include "hangcheck-timer.h"

static inline unsigned long long get_clock (void)
{
        unsigned long long clk;

        __asm__("stck 0(%0)" : : "a" (&(clk)) : "memory", "cc");
        return clk;
}

#define DEFAULT_IOFENCE_MARGIN 60	/* Default fudge factor, in seconds */
#define DEFAULT_IOFENCE_TICK 180	/* Default timer timeout, in seconds */

static int hangcheck_tick = DEFAULT_IOFENCE_TICK;
static int hangcheck_margin = DEFAULT_IOFENCE_MARGIN;

MODULE_PARM(hangcheck_tick,"i");
MODULE_PARM_DESC(hangcheck_tick, "Timer delay.");
MODULE_PARM(hangcheck_margin,"i");
MODULE_PARM_DESC(hangcheck_margin, "If the hangcheck timer has been delayed more than hangcheck_margin seconds, the machine will reboot.");


static void hangcheck_fire(unsigned long);

/* Last time scheduled */
static unsigned long long hangcheck_clock, hangcheck_clock_margin;

static struct timer_list hangcheck_ticktock = {
	function:	hangcheck_fire,
};


static void hangcheck_fire(unsigned long data)
{
	unsigned long long cur_clock, clock_diff;

	cur_clock = get_clock();

	if (cur_clock > hangcheck_clock)
		clock_diff = cur_clock - hangcheck_clock;
	else
		clock_diff = (cur_clock + (~0ULL - hangcheck_clock)); /* or something */

#if 0
	printk(KERN_CRIT "cur_clock = %lx.%lx, hangcheck_clock diff is %lx.%lx.\n",
		(cur_clock >> 32) & 0xFFFFFFFFULL, cur_clock & 0xFFFFFFFFULL,
		(hangcheck_clock >> 32) & 0xFFFFFFFFULL, hangcheck_clock & 0xFFFFFFFFULL );
	printk(KERN_CRIT "hangcheck_margin = %lu, HZ = %lu.\n", hangcheck_margin, HZ);
	printk(KERN_CRIT "clock_diff = %lu.%lu, predicted diff is %lu.%lu.\n",
	       (unsigned long) ((clock_diff >> 32) & 0xFFFFFFFFULL),
	       (unsigned long) (clock_diff & 0xFFFFFFFFULL),
	       (unsigned long) ((hangcheck_clock_margin >> 32) & 0xFFFFFFFFULL),
	       (unsigned long) (hangcheck_clock_margin & 0xFFFFFFFFULL));
	printk(KERN_CRIT "hangcheck_margin = %lu, HZ = %lu.\n", hangcheck_margin, HZ);
#endif
	
	if (clock_diff > hangcheck_clock_margin) {
		printk(KERN_CRIT "hangcheck is restarting the machine.\n");
#if 	1
		machine_restart(NULL);
#endif
	}
	mod_timer(&hangcheck_ticktock, jiffies + (hangcheck_tick*HZ));
	hangcheck_clock = get_clock();
}  /* hangcheck_fire() */


static int __init hangcheck_init(void)
{
        version_hash_print();
	printk("Starting hangcheck timer (tick is %d seconds, margin is %d seconds).\n",
	       hangcheck_tick, hangcheck_margin);

	/* FA240000 is 1 Second in the IBM time universe (Page 4-38 Principles of Op for zSeries) */
	hangcheck_clock_margin = (unsigned long long)(hangcheck_margin + hangcheck_tick) * 0xFA240000ULL;

	hangcheck_clock = get_clock();
	mod_timer(&hangcheck_ticktock, jiffies + (hangcheck_tick*HZ));

	return 0;
}  /* hangcheck_init() */


static void __exit hangcheck_exit(void)
{
	printk("Stopping hangcheck timer.\n");

	lock_kernel();
	del_timer(&hangcheck_ticktock);
	unlock_kernel();
}  /* hangcheck_exit() */

module_init(hangcheck_init);
module_exit(hangcheck_exit);
