/*
 * originally based on the dummy device.
 *
 * Copyright 1999, Thomas Davis, tadavis@lbl.gov.  
 * Licensed under the GPL. Based on dummy.c, and eql.c devices.
 *
 * bonding.c: an Ethernet Bonding driver
 *
 * This is useful to talk to a Cisco EtherChannel compatible equipment:
 *	Cisco 5500
 *	Sun Trunking (Solaris)
 *	Alteon AceDirector Trunks
 *	Linux Bonding
 *	and probably many L2 switches ...
 *
 * How it works:
 *    ifconfig bond0 ipaddress netmask up
 *      will setup a network device, with an ip address.  No mac address 
 *	will be assigned at this time.  The hw mac address will come from 
 *	the first slave bonded to the channel.  All slaves will then use 
 *	this hw mac address.
 *
 *    ifconfig bond0 down
 *         will release all slaves, marking them as down.
 *
 *    ifenslave bond0 eth0
 *	will attach eth0 to bond0 as a slave.  eth0 hw mac address will either
 *	a: be used as initial mac address
 *	b: if a hw mac address already is there, eth0's hw mac address 
 *	   will then be set from bond0.
 *
 * v0.1 - first working version.
 * v0.2 - changed stats to be calculated by summing slaves stats.
 *
 * Changes:
 * Arnaldo Carvalho de Melo <acme@conectiva.com.br>
 * - fix leaks on failure at bond_init
 *
 * 2000/09/30 - Willy Tarreau <willy at meta-x.org>
 *     - added trivial code to release a slave device.
 *     - fixed security bug (CAP_NET_ADMIN not checked)
 *     - implemented MII link monitoring to disable dead links :
 *       All MII capable slaves are checked every <miimon> milliseconds
 *       (100 ms seems good). This value can be changed by passing it to
 *       insmod. A value of zero disables the monitoring (default).
 *     - fixed an infinite loop in bond_xmit_roundrobin() when there's no
 *       good slave.
 *     - made the code hopefully SMP safe
 *
 * 2000/10/03 - Willy Tarreau <willy at meta-x.org>
 *     - optimized slave lists based on relevant suggestions from Thomas Davis
 *     - implemented active-backup method to obtain HA with two switches:
 *       stay as long as possible on the same active interface, while we
 *       also monitor the backup one (MII link status) because we want to know
 *       if we are able to switch at any time. ( pass "mode=1" to insmod )
 *     - lots of stress testings because we need it to be more robust than the
 *       wires ! :->
 *
 * 2000/10/09 - Willy Tarreau <willy at meta-x.org>
 *     - added up and down delays after link state change.
 *     - optimized the slaves chaining so that when we run forward, we never
 *       repass through the bond itself, but we can find it by searching
 *       backwards. Renders the deletion more difficult, but accelerates the
 *       scan.
 *     - smarter enslaving and releasing.
 *     - finer and more robust SMP locking
 *
 * 2000/10/17 - Willy Tarreau <willy at meta-x.org>
 *     - fixed two potential SMP race conditions
 *
 * 2000/10/18 - Willy Tarreau <willy at meta-x.org>
 *     - small fixes to the monitoring FSM in case of zero delays
 * 2000/11/01 - Willy Tarreau <willy at meta-x.org>
 *     - fixed first slave not automatically used in trunk mode.
 * 2000/11/10 : spelling of "EtherChannel" corrected.
 * 2000/11/13 : fixed a race condition in case of concurrent accesses to ioctl().
 * 2000/12/16 : fixed improper usage of rtnl_exlock_nowait().
 *
 * 2001/1/3 - Chad N. Tindel <ctindel at ieee dot org>
 *     - The bonding driver now simulates MII status monitoring, just like
 *       a normal network device.  It will show that the link is down iff
 *       every slave in the bond shows that their links are down.  If at least
 *       one slave is up, the bond's MII status will appear as up.
 *
 * 2001/2/7 - Chad N. Tindel <ctindel at ieee dot org>
 *     - Applications can now query the bond from user space to get
 *       information which may be useful.  They do this by calling
 *       the BOND_INFO_QUERY ioctl.  Once the app knows how many slaves
 *       are in the bond, it can call the BOND_SLAVE_INFO_QUERY ioctl to
 *       get slave specific information (# link failures, etc).  See
 *       <linux/if_bonding.h> for more details.  The structs of interest
 *       are ifbond and ifslave.
 *
 * 2001/4/5 - Chad N. Tindel <ctindel at ieee dot org>
 *     - Ported to 2.4 Kernel
 * 
 * 2001/5/2 - Jeffrey E. Mast <jeff at mastfamily dot com>
 *     - When a device is detached from a bond, the slave device is no longer
 *       left thinking that is has a master.
 *
 * 2001/5/16 - Jeffrey E. Mast <jeff at mastfamily dot com>
 *     - memset did not appropriately initialized the bond rw_locks. Used 
 *       rwlock_init to initialize to unlocked state to prevent deadlock when 
 *       first attempting a lock
 *     - Called SET_MODULE_OWNER for bond device
 *
 * 2001/5/17 - Tim Anderson <tsa at mvista.com>
 *     - 2 paths for releasing for slave release; 1 through ioctl
 *       and 2) through close. Both paths need to release the same way.
 *     - the free slave in bond release is changing slave status before
 *       the free. The netdev_set_master() is intended to change slave state
 *       so it should not be done as part of the release process.
 *     - Simple rule for slave state at release: only the active in A/B and
 *       only one in the trunked case.
 *
 * 2001/6/01 - Tim Anderson <tsa at mvista.com>
 *     - Now call dev_close when releasing a slave so it doesn't screw up
 *       out routing table.
 *
 * 2001/6/01 - Chad N. Tindel <ctindel at ieee dot org>
 *     - Added /proc support for getting bond and slave information.
 *       Information is in /proc/net/<bond device>/info. 
 *     - Changed the locking when calling bond_close to prevent deadlock.
 */

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/types.h>
#include <linux/fcntl.h>
#include <linux/interrupt.h>
#include <linux/ptrace.h>
#include <linux/ioport.h>
#include <linux/in.h>
#include <linux/malloc.h>
#include <linux/string.h>
#include <linux/init.h>
#include <linux/timer.h>
#include <linux/socket.h>
#include <asm/system.h>
#include <asm/bitops.h>
#include <asm/io.h>
#include <asm/dma.h>
#include <asm/uaccess.h>
#include <linux/errno.h>

#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>
#include <net/sock.h>
#include <linux/rtnetlink.h>

#include <linux/if_bonding.h>
#include <linux/smp.h>

/* monitor all links that often (in milliseconds). <=0 disables monitoring */
#ifndef BOND_LINK_MON_INTERV
#define BOND_LINK_MON_INTERV	0
#endif

#undef  MII_LINK_UP
#define MII_LINK_UP	0x04

#undef  MII_ENDOF_NWAY
#define MII_ENDOF_NWAY	0x20

#undef  MII_LINK_READY
#define MII_LINK_READY	(MII_LINK_UP | MII_ENDOF_NWAY)

#define MAX_BOND_ADDR 256

static int miimon    = BOND_LINK_MON_INTERV;
static int mode      = BOND_MODE_ROUNDROBIN;
static int updelay   = 0;
static int downdelay = 0;

MODULE_PARM(miimon, "i");
MODULE_PARM_DESC(miimon, "Link check interval in milliseconds");
MODULE_PARM(mode, "i");
MODULE_PARM_DESC(mode, "Mode of operation : 0 for round robin, 1 for active-backup");
MODULE_PARM(updelay, "i");
MODULE_PARM_DESC(updelay, "Delay before considering link up, in milliseconds");
MODULE_PARM(downdelay, "i");
MODULE_PARM_DESC(downdelay, "Delay before considering link down, in milliseconds");

static int bond_xmit_roundrobin(struct sk_buff *skb, struct net_device *dev);
static int bond_xmit_activebackup(struct sk_buff *skb, struct net_device *dev);
static struct net_device_stats *bond_get_stats(struct net_device *dev);
static void bond_monitor(struct net_device *dev);
static int bond_release(struct net_device *master, struct net_device *slave);

/*
 * bond_get_info is the interface into the /proc filesystem.  This is
 * a different interface than the BOND_INFO_QUERY ioctl.  That is done
 * through the generic networking ioctl interface, and bond_info_query
 * is the internal function which provides that information.
 */
static int bond_get_info(char *buf, char **start, off_t offset, int length);

static struct net_device *this_bond = NULL;

/* #define BONDING_DEBUG 1 */

/* several macros */

/*#define IS_UP(dev)	(((dev)->flags & (IFF_UP | IFF_RUNNING)) == \
                         (IFF_UP | IFF_RUNNING))
 */
#define IS_UP(dev)	((((dev)->flags & (IFF_UP)) == (IFF_UP)) && \
                     (netif_running(dev) && netif_carrier_ok(dev)))

/* 
 * This function detaches the slave <slave> from the list <bond>.
 * WARNING: no check is made to verify if the slave effectively
 * belongs to <bond>. It returns <slave> in case it's needed.
 * Nothing is freed on return, structures are just unchained.
 * If the bond->current_slave pointer was pointing to <slave>,
 * it's replaced with slave->next, or <bond> if not applicable.
 */
static slave_t *bond_detach_slave(bonding_t *bond, slave_t *slave)
{
	if ((bond == NULL) || (slave == NULL) ||
	    ((void *)bond == (void *)slave)) {
		printk(KERN_ERR
		       "bond_detach_slave(): trying to detach "
		       "slave %p from bond %p\n", bond, slave);
		return slave;
	}

	if (bond->next == slave) {  /* is the slave at the head ? */
		if (bond->prev == slave) {  /* is the slave alone ? */
			write_lock(&bond->ptrlock);
			bond->current_slave = NULL; /* no slave anymore */
			write_unlock(&bond->ptrlock);
			bond->prev = bond->next = (slave_t *)bond;
		}
		else { /* not alone */
			bond->next        = slave->next;
			slave->next->prev = (slave_t *)bond;
			bond->prev->next  = slave->next;

			write_lock(&bond->ptrlock);
			if (bond->current_slave == slave)
				bond->current_slave = slave->next;
			write_unlock(&bond->ptrlock);
		}
	}
	else {
		slave->prev->next = slave->next;
		if (bond->prev == slave)  /* is this slave the last one ? */
			bond->prev = slave->prev;
		else
			slave->next->prev = slave->prev;

		write_lock(&bond->ptrlock);
		if (bond->current_slave == slave)
			bond->current_slave = slave->next;
		write_unlock(&bond->ptrlock);
	}

	return slave;
}

/* 
 * if <dev> supports MII link status reporting, check its link
 * and report it as a bit field in a short int :
 *   - 0x04 means link is up,
 *   - 0x20 means end of autonegociation
 * If the device doesn't support MII, then we only report 0x24,
 * meaning that the link is up and running since we can't check it.
 */
static u16 bond_check_dev_link(struct net_device *dev)
{
	static int (* ioctl)(struct net_device *, struct ifreq *, int);
	struct ifreq ifr;
	u16 *data = (u16 *)&ifr.ifr_data;
		
	/* data[0] automagically filled by the ioctl */
	data[1] = 1; /* MII location 1 reports Link Status */

	if (((ioctl = dev->do_ioctl) != NULL) &&  /* ioctl to access MII */
	    (ioctl(dev, &ifr, SIOCDEVPRIVATE) == 0)) {
		/* now, data[3] contains info about link status :
		   - data[3] & 0x04 means link up
		   - data[3] & 0x20 means end of auto-negociation
		*/
		return data[3];
	}
	else {
		return MII_LINK_READY;  /* spoof link up ( we can't check it) */
	}
}

static u16 bond_check_mii_link(struct net_device *master)
{
	bonding_t *bond = (struct bonding *)master->priv;
	int has_active_interface = 0;

	read_lock(&bond->ptrlock);
	has_active_interface = (bond->current_slave != NULL);
	read_unlock(&bond->ptrlock);

	return (has_active_interface ? MII_LINK_READY : 0);
}

static int bond_open(struct net_device *dev)
{
	struct timer_list *timer = &((struct bonding *)(dev->priv))->timer;
	MOD_INC_USE_COUNT;

	if (miimon > 0) {  /* link check interval, in milliseconds. */
		init_timer(timer);
		timer->expires  = jiffies + (miimon * HZ / 1000);
		timer->data     = (unsigned long)dev;
		timer->function = (void *)&bond_monitor;
		add_timer(timer);
	}

	return 0;
}

static int bond_close(struct net_device *master)
{
	bonding_t *bond = (struct bonding *) master->priv;
	slave_t *slave;
	unsigned long flags;

	write_lock_irqsave(&bond->lock, flags);

	if (miimon > 0) {  /* link check interval, in milliseconds. */
		del_timer(&bond->timer);
	}
	/* We need to unlock this because bond_release will re-lock it */
	write_unlock_irqrestore(&bond->lock, flags);

	/* Release the bonded slaves */
	while ((slave = bond->prev) != (slave_t *)bond) {
		bond_release(master, slave->dev);
	}

	MOD_DEC_USE_COUNT;
	return 0;
}

static void set_multicast_list(struct net_device *master)
{
/*
	bonding_t *bond = master->priv;
	slave_t *slave;

	for (slave = bond->next; slave != (slave_t*)bond; slave = slave->next) {
		slave->dev->mc_list = master->mc_list;
		slave->dev->mc_count = master->mc_count;
		slave->dev->flags = master->flags;
		slave->dev->set_multicast_list(slave->dev);
	}
 */
}

/* enslave device <slave> to bond device <master> */
static int bond_enslave(struct net_device *master_dev, 
                        struct net_device *slave_dev)
{
	bonding_t *bond;
	slave_t *new_slave;
	unsigned long flags;
	int err;
	
	if (master_dev == NULL || slave_dev == NULL) 
		return -ENODEV;

	bond = (struct bonding *) master_dev->priv;

	if (slave_dev->do_ioctl == NULL) {
		printk(KERN_DEBUG
		       "Warning : no link monitoring support for %s\n",
		       slave_dev->name);
	}
	write_lock_irqsave(&bond->lock, flags);

	/* not running. */
	if ((slave_dev->flags & IFF_UP) != IFF_UP) {
#ifdef BONDING_DEBUG
		printk(KERN_CRIT "Error, slave_dev is not running\n");
#endif
		write_unlock_irqrestore(&bond->lock, flags);
		return -EINVAL;
	}

	/* already enslaved */
	if (master_dev->flags & IFF_SLAVE || slave_dev->flags & IFF_SLAVE) {
#ifdef BONDING_DEBUG
		printk(KERN_CRIT "Error, Device was already enslaved\n");
#endif
		write_unlock_irqrestore(&bond->lock, flags);
		return -EBUSY;
	}
		   
	if ((new_slave = kmalloc(sizeof(slave_t), GFP_KERNEL)) == NULL) {
		write_unlock_irqrestore(&bond->lock, flags);
		return -ENOMEM;
	}
	memset(new_slave, 0, sizeof(slave_t));

	err = netdev_set_master(slave_dev, master_dev);

	if (err) {
#ifdef BONDING_DEBUG
		printk(KERN_CRIT "Error %d calling netdev_set_master\n", err);
#endif
		kfree(new_slave);
		return err;      
	}

	new_slave->dev = slave_dev;

	/* queue to the end of the slaves list, make the first element its
	   successor, the last one its predecessor, and make it the bond's
	   predecessor. */
	new_slave->prev       = bond->prev;
	new_slave->prev->next = new_slave;
	bond->prev            = new_slave;
	new_slave->next       = bond->next;

	new_slave->delay = 0;
	new_slave->link_failure_count = 0;

	/* check for initial state */
	if ((miimon <= 0) ||
	    ((bond_check_dev_link(slave_dev) & MII_LINK_READY) == MII_LINK_READY)) {
#ifdef BONDING_DEBUG
		printk(KERN_CRIT "Initial state of slave_dev is BOND_LINK_UP\n");
#endif
		new_slave->link  = BOND_LINK_UP;
	}
	else {
#ifdef BONDING_DEBUG
		printk(KERN_CRIT "Initial state of slave_dev is BOND_LINK_DOWN\n");
#endif
		new_slave->link  = BOND_LINK_DOWN;
	}

	/* if we're in active-backup mode, we need one and only one active
	 interface. The backup interfaces will have their NOARP flag set
	 because we need them to be completely deaf and not to respond to
	 any ARP request on the network to avoid fooling a switch. Thus,
	 since we guarantee that current_slave always point to the last
	 usable interface, we just have to verify this interface's flag.
	*/
	if (mode == BOND_MODE_ACTIVEBACKUP) {
		if (((bond->current_slave == NULL)
		     || (bond->current_slave->dev->flags & IFF_NOARP))
		    && (new_slave->link == BOND_LINK_UP)) {
#ifdef BONDING_DEBUG
			printk(KERN_CRIT "This is the first active slave\n");
#endif
			/* first slave or no active slave yet, and this link
			   is OK, so make this interface the active one */
			bond->current_slave = new_slave;
			new_slave->state = BOND_STATE_ACTIVE;
			new_slave->dev->flags &= ~IFF_NOARP; /* active interface */
		}
		else {
#ifdef BONDING_DEBUG
			printk(KERN_CRIT "This is just a backup slave\n");
#endif
			new_slave->dev->flags |= IFF_NOARP; /* backup interface */
			new_slave->state = BOND_STATE_BACKUP;
		}
	} else {
#ifdef BONDING_DEBUG
		printk(KERN_CRIT "This slave is always active in trunk mode\n");
#endif
		/* always active in trunk mode */
		new_slave->state = BOND_STATE_ACTIVE;
		if (bond->current_slave == NULL)
			bond->current_slave = new_slave;
	}

	write_unlock_irqrestore(&bond->lock, flags);

	printk (KERN_INFO "%s: enslaving %s as a%s interface with a%s link.\n",
		master_dev->name, slave_dev->name,
		new_slave->state == BOND_STATE_ACTIVE ? "n active" : " backup",
		new_slave->link == BOND_LINK_UP ? "n up" : " down");

	return 0;
}

/* Choose a new valid interface from the pool, set it active
   and make it the current slave. If no valid interface is
   found, the oldest slave in BACK state is choosen and
   activated. If none is found, it's considered as no
   interfaces left so the current slave is set to NULL.
   The result is a pointer to the current slave.

   Since this function sends messages tails through printk, the caller
   must have started something like `printk(KERN_INFO "xxxx ");'.

   Warning: must put locks around the call to this function if needed.
*/
slave_t *change_active_interface(bonding_t *bond)
{
	slave_t *newslave, *oldslave;
	slave_t *bestslave = NULL;
	int mintime;

	read_lock(&bond->ptrlock);
	newslave = oldslave = bond->current_slave;
	read_unlock(&bond->ptrlock);

	if (newslave == NULL) { /* there were no active slaves left */
		if (bond->next != (slave_t *)bond) {  /* found one slave */
			write_lock(&bond->ptrlock);
			newslave = bond->current_slave = bond->next;
			write_unlock(&bond->ptrlock);
		}
		else {
			printk (" but could not find any %s interface.\n",
				(mode == BOND_MODE_ACTIVEBACKUP) ? "backup":"other");
			write_lock(&bond->ptrlock);
			bond->current_slave = (slave_t *)NULL;
			write_unlock(&bond->ptrlock);
			return NULL; /* still no slave, return NULL */
		}
	}

	mintime = updelay;

	do {
		if (IS_UP(newslave->dev)) {
			if (newslave->link == BOND_LINK_UP) {
				/* this one is immediately usable */
				if (mode == BOND_MODE_ACTIVEBACKUP) {
					newslave->state = BOND_STATE_ACTIVE;
					newslave->dev->flags &= ~IFF_NOARP;

					printk (" and making interface %s the active one.\n",
						newslave->dev->name);
				}
				else
					printk (" and setting pointer to interface %s.\n",
						newslave->dev->name);

				write_lock(&bond->ptrlock);
				bond->current_slave = newslave;
				write_unlock(&bond->ptrlock);
				return newslave;
			}
			else if (newslave->link == BOND_LINK_BACK) {
				/* link up, but waiting for stabilization */
				if (newslave->delay < mintime) {
					mintime = newslave->delay;
					bestslave = newslave;
				}
			}
		}
	} while ((newslave = newslave->next) != oldslave);

	/* no usable backup found, we'll see if we at least got a link that was
	   coming back for a long time, and could possibly already be usable.
	*/

	if (bestslave != NULL) {
		/* early take-over. */
		printk (" and making interface %s the active one %d ms earlier.\n",
			bestslave->dev->name,
			(updelay - bestslave->delay)*miimon);

		bestslave->delay = 0;
		bestslave->link = BOND_LINK_UP;
		bestslave->state = BOND_STATE_ACTIVE;
		bestslave->dev->flags &= ~IFF_NOARP;

		write_lock(&bond->ptrlock);
		bond->current_slave = bestslave;
		write_unlock(&bond->ptrlock);
		return bestslave;
	}

	printk (" but could not find any %s interface.\n",
		(mode == BOND_MODE_ACTIVEBACKUP) ? "backup":"other");
	
	/* absolutely nothing found. let's return NULL */
	write_lock(&bond->ptrlock);
	bond->current_slave = (slave_t *)NULL;
	write_unlock(&bond->ptrlock);
	return NULL;
}

/*
 * Try to release the slave device <slave> from the bond device <master>
 * It is legal to access current_slave without a lock because all the function
 * is write-locked.
 *
 * The rules for slave state should be:
 *   for Active/Backup:
 *     Active stays on all backups go down
 *   for Bonded connections:
 *     The first up interface should be left on and all others downed.
 */
static int bond_release(struct net_device *master, struct net_device *slave)
{
	bonding_t *bond;
	slave_t *our_slave, *old_current;
	unsigned long flags;
	
	if (master == NULL || slave == NULL) 
		return -ENODEV;

	bond = (struct bonding *) master->priv;

	write_lock_irqsave(&bond->lock, flags);

	/* master already enslaved, or slave not enslaved,
	   or no slave for this master */
	if ((master->flags & IFF_SLAVE) || !(slave->flags & IFF_SLAVE)) {
		printk (KERN_DEBUG "%s: cannot release %s.\n", master->name, slave->name);
		write_unlock_irqrestore(&bond->lock, flags);
		return -EINVAL;
	}

	our_slave = (slave_t *)bond;
	old_current = bond->current_slave;
	while ((our_slave = our_slave->prev) != (slave_t *)bond) {
		if (our_slave->dev == slave) {
			bond_detach_slave(bond, our_slave);

			printk (KERN_INFO "%s: releasing %s interface %s",
				master->name,
				(our_slave->state == BOND_STATE_ACTIVE) ? "active" : "backup",
				slave->name);

			if (our_slave == old_current)
				/* find a new interface and be verbose */
				change_active_interface(bond); 
			else
				printk(".\n");
			kfree(our_slave);

			/* release the slave from its bond */
			 
			netdev_set_master(slave, NULL);

			/* only restore its RUNNING flag if monitoring set it down */
			if (slave->flags & IFF_UP)
				slave->flags |= IFF_RUNNING;

			if (slave->flags & IFF_NOARP || bond->current_slave != NULL)
				dev_close(slave);

			if (bond->current_slave == NULL) {
				printk(KERN_INFO
				       "%s: now running without any active interface !\n",
				       master->name);
			}

			write_unlock_irqrestore(&bond->lock, flags);
			return 0;  /* deletion OK */
		}
	}

	/* if we get here, it's because the device was not found */
	write_unlock_irqrestore(&bond->lock, flags);

	printk (KERN_INFO "%s: %s not enslaved\n", master->name, slave->name);
	return -EINVAL;
}

/* this function is called regularly to monitor each slave's link. */
static void bond_monitor(struct net_device *master)
{
	bonding_t *bond = (struct bonding *) master->priv;
	slave_t *slave, *bestslave, *oldcurrent;
	unsigned long flags;
	int slave_died = 0;

	read_lock_irqsave(&bond->lock, flags);

	if (rtnl_shlock_nowait() == -EAGAIN) {
		goto monitor_out;
	}

	if (rtnl_exlock_nowait() == -EAGAIN) {
		rtnl_shunlock();
		goto monitor_out;
	}
	/* we will try to read the link status of each of our slaves, and
	   set their IFF_RUNNING flag appropriately. For each slave not
	   supporting MII status, we won't do anything so that a user-space
	   program could monitor the link itself if needed.
	*/

	bestslave = NULL;
	slave = (slave_t *)bond;

	read_lock(&bond->ptrlock);
	oldcurrent = bond->current_slave;
	read_unlock(&bond->ptrlock);

	while ((slave = slave->prev) != (slave_t *)bond) {
		/* use updelay+1 to match an UP slave even when updelay is 0 */
		int mindelay = updelay + 1;
		struct net_device *dev = slave->dev;
		u16 link_state;
		
		link_state = bond_check_dev_link(dev);

		switch (slave->link) {
		case BOND_LINK_UP:	/* the link was up */
			if ((link_state & MII_LINK_UP) == MII_LINK_UP) {
				/* link stays up, tell that this one
				   is immediately available */
				if (IS_UP(dev) && (mindelay > -2)) {
					/* -2 is the best case :
					   this slave was already up */
					mindelay = -2;
					bestslave = slave;
				}
				break;
			}
			else { /* link going down */
				slave->link  = BOND_LINK_FAIL;
				slave->delay = downdelay;
				if (slave->link_failure_count < UINT_MAX) {
					slave->link_failure_count++;
				}
				if (downdelay > 0)
					printk (KERN_INFO
						"%s: link status down for %sinterface "
						"%s, disabling it in %d ms.\n",
						master->name,
						IS_UP(dev)
						? ((mode == BOND_MODE_ACTIVEBACKUP)
						   ? ((slave == oldcurrent)
						      ? "active " : "backup ")
						   : "")
						: "idle ",
						dev->name,
						downdelay * miimon);
			}
			/* no break ! fall through the BOND_LINK_FAIL test to
			   ensure proper action to be taken
			*/
		case BOND_LINK_FAIL:	/* the link has just gone down */
			if ((link_state & MII_LINK_UP) == 0) {
				/* link stays down */
				if (slave->delay <= 0) {
					/* link down for too long time */
					slave->link = BOND_LINK_DOWN;
					/* in active/backup mode, we must
					   completely disable this interface */
					if (mode == BOND_MODE_ACTIVEBACKUP) {
						slave->state = BOND_STATE_BACKUP;
						dev->flags |= IFF_NOARP;
					}
					printk(KERN_INFO
					       "%s: link status definitely down "
					       "for interface %s, disabling it",
					       master->name,
					       dev->name);

					read_lock(&bond->ptrlock);
					if (slave == bond->current_slave) {
						read_unlock(&bond->ptrlock);
						/* find a new interface and be verbose */
						change_active_interface(bond);
					}
					else {
						read_unlock(&bond->ptrlock);
						printk(".\n");
					}
					slave_died = 1;
				}
				else
					slave->delay--;
			}
			else if ((link_state & MII_LINK_READY) == MII_LINK_READY) {
				/* link up again */
				slave->link  = BOND_LINK_UP;
				printk(KERN_INFO
				       "%s: link status up again after %d ms "
				       "for interface %s.\n",
				       master->name,
				       (downdelay - slave->delay) * miimon,
				       dev->name);

				if (IS_UP(dev) && (mindelay > -1)) {
					/* -1 is a good case : this slave went
					   down only for a short time */
					mindelay = -1;
					bestslave = slave;
				}
			}
			break;
		case BOND_LINK_DOWN:	/* the link was down */
			if ((link_state & MII_LINK_READY) != MII_LINK_READY)
				/* the link stays down, nothing more to do */
				break;
			else {	/* link going up */
				slave->link  = BOND_LINK_BACK;
				slave->delay = updelay;
				
				if (updelay > 0)
					/* if updelay == 0, no need to
					   advertise about a 0 ms delay */
					printk (KERN_INFO
						"%s: link status up for interface"
						" %s, enabling it in %d ms.\n",
						master->name,
						dev->name,
						updelay * miimon);
			}
			/* no break ! fall through the BOND_LINK_BACK state in
			   case there's something to do.
			*/
		case BOND_LINK_BACK:	/* the link has just come back */
			if ((link_state & MII_LINK_UP) == 0) {
				/* link down again */
				slave->link  = BOND_LINK_DOWN;
				printk(KERN_INFO
				       "%s: link status down again after %d ms "
				       "for interface %s.\n",
				       master->name,
				       (updelay - slave->delay) * miimon,
				       dev->name);
			}
			else if ((link_state & MII_LINK_READY) == MII_LINK_READY) {
				/* link stays up */
				if (slave->delay == 0) {
					/* now the link has been up for long time enough */
					slave->link = BOND_LINK_UP;

					if (mode == BOND_MODE_ACTIVEBACKUP) {
						/* prevent it from being the active one */
						slave->state = BOND_STATE_BACKUP;
					}
					else {
						/* make it immediately active */
						slave->state = BOND_STATE_ACTIVE;
					}

					printk(KERN_INFO
					       "%s: link status definitely up "
					       "for interface %s.\n",
					       master->name,
					       dev->name);
				}
				else
					slave->delay--;
				
				/* we'll also look for the mostly eligible slave */
				if (IS_UP(dev) && (slave->delay < mindelay)) {
					mindelay = slave->delay;
					bestslave = slave;
				} 
			}
			break;
		} /* end of switch */
	} /* end of while */

	/* if there's no active interface and we discovered that one
	   of the slaves could be activated earlier, so we do it.
	*/
	read_lock(&bond->ptrlock);
	oldcurrent=bond->current_slave;
	read_unlock(&bond->ptrlock);

	if (oldcurrent == NULL) {  /* no active interface at the moment */
		if (bestslave != NULL) { /* last chance to find one ? */
			if (bestslave->link == BOND_LINK_UP)
				printk (KERN_INFO
					"%s: making interface %s the new active one.\n",
					master->name, bestslave->dev->name);
			else {
				printk (KERN_INFO
					"%s: making interface %s the new "
					"active one %d ms earlier.\n",
					master->name, bestslave->dev->name,
					(updelay - bestslave->delay) * miimon);

				bestslave->delay= 0;
				bestslave->link = BOND_LINK_UP;
			}
			bestslave->state = BOND_STATE_ACTIVE;

			if (mode == BOND_MODE_ACTIVEBACKUP)
				bestslave->dev->flags &= ~IFF_NOARP;

			write_lock(&bond->ptrlock);
			bond->current_slave = bestslave;
			write_unlock(&bond->ptrlock);
		}
		else if (slave_died) {
			/* print this message only once a slave has just died */
			printk(KERN_INFO
			       "%s: now running without any active interface !\n",
			       master->name);
		}
	}

	rtnl_exunlock();
	rtnl_shunlock();
monitor_out:
	read_unlock_irqrestore(&bond->lock, flags);
	/* re-arm the timer */
	mod_timer(&bond->timer, jiffies + (miimon * HZ / 1000));
}

static int bond_sethwaddr(struct net_device *master, struct net_device *slave)
{
	memcpy(master->dev_addr, slave->dev_addr, slave->addr_len);
	return 0;
}

static int bond_info_query(struct net_device *master, struct ifbond *info)
{
	bonding_t *bond = (struct bonding *) master->priv;
	slave_t *slave;

	info->bond_mode = mode;
	info->num_slaves = 0;
	info->miimon = miimon;

	read_lock(&bond->ptrlock);
	for (slave = bond->prev; slave != (slave_t *)bond; slave = slave->prev) {
		info->num_slaves++;
	}
	read_unlock(&bond->ptrlock);

	return 0;
}

static int bond_slave_info_query(struct net_device *master, 
                                 struct ifslave *info)
{
	bonding_t *bond = (struct bonding *) master->priv;
	slave_t *slave;
	int cur_ndx = 0;

	if (info->slave_id < 0) {
		return -ENODEV;
	}

	read_lock(&bond->ptrlock);
	for (slave = bond->prev; 
		 slave != (slave_t *)bond && cur_ndx < info->slave_id; 
		 slave = slave->prev) {
		cur_ndx++;
	}
	if (cur_ndx == info->slave_id) {
		strcpy(info->slave_name, slave->dev->name);
		info->link = slave->link;
		info->state = slave->state;
		info->link_failure_count = slave->link_failure_count;
	} else {
		read_unlock(&bond->ptrlock);
		return -ENODEV;
	}
	read_unlock(&bond->ptrlock);

	return 0;
}

static int bond_ioctl(struct net_device *master_dev, struct ifreq *ifr, int cmd)
{
	struct net_device *slave_dev = dev_get_by_name(ifr->ifr_slave);
	struct ifbond *u_binfo = NULL, k_binfo;
	struct ifslave *u_sinfo = NULL, k_sinfo;
    u16 *data = NULL;
	int ret = 0;

#ifdef BONDING_DEBUG
	printk("master=%s, cmd=%d\n", master_dev->name, cmd);
#endif

	switch (cmd) {
	case SIOCDEVPRIVATE:
		data = (u16 *)&ifr->ifr_data;
		if (data == NULL) {
			return -EINVAL;
		}
		data[0] = 0;
		/* Fall Through */
	case SIOCDEVPRIVATE + 1:
		/* 
		 * We do this again just in case we were called by SIOCDEVPRIVATE + 1
		 * instead of SIOCDEVPRIVATE.
		 */
		data = (u16 *)&ifr->ifr_data;
		if (data == NULL) {
			return -EINVAL;
		}
		if (data[1] == 1) {
			data[3] = bond_check_mii_link(master_dev);
		}
		return 0;
    case BOND_INFO_QUERY:
		u_binfo = (struct ifbond *)ifr->ifr_data;
		if (copy_from_user(&k_binfo, u_binfo, sizeof(ifbond))) {
			return -EFAULT;
		}
		ret = bond_info_query(master_dev, &k_binfo);
		if (ret == 0) {
			if (copy_to_user(u_binfo, &k_binfo, sizeof(ifbond))) {
				return -EFAULT;
			}
		}
		return ret;
    case BOND_SLAVE_INFO_QUERY:
		u_sinfo = (struct ifslave *)ifr->ifr_data;
		if (copy_from_user(&k_sinfo, u_sinfo, sizeof(ifslave))) {
			return -EFAULT;
		}
		ret = bond_slave_info_query(master_dev, &k_sinfo);
		if (ret == 0) {
			if (copy_to_user(u_sinfo, &k_sinfo, sizeof(ifslave))) {
				return -EFAULT;
			}
		}
		return ret;

	}

	if (!capable(CAP_NET_ADMIN)) {
		return -EPERM;
	}

	switch (cmd) {
	case BOND_ENSLAVE:
		return bond_enslave(master_dev, slave_dev);
	case BOND_RELEASE:
		return bond_release(master_dev, slave_dev);
	case BOND_SETHWADDR:
		return bond_sethwaddr(master_dev, slave_dev);
	default:
		return -EOPNOTSUPP;
	}
}

#ifdef CONFIG_NET_FASTROUTE
static int bond_accept_fastpath(struct net_device *dev, struct dst_entry *dst)
{
	return -1;
}
#endif

static int bond_xmit_roundrobin(struct sk_buff *skb, struct net_device *dev)
{
	slave_t *slave, *start_at;
	struct bonding *bond = (struct bonding *) dev->priv;
	unsigned long flags;

	if (!IS_UP(dev)) { /* bond down */
		dev_kfree_skb(skb);
		return 0;
	}

	read_lock_irqsave(&bond->lock, flags);

	read_lock(&bond->ptrlock);
	slave = start_at = bond->current_slave;
	read_unlock(&bond->ptrlock);

	if (slave == NULL) { /* we're at the root, get the first slave */
		/* no suitable interface, frame not sent */
		dev_kfree_skb(skb);
		read_unlock_irqrestore(&bond->lock, flags);
		return 0;
	}

	do {
		if (IS_UP(slave->dev)
		    && (slave->link == BOND_LINK_UP)
		    && (slave->state == BOND_STATE_ACTIVE)) {

			skb->dev = slave->dev;
			skb->priority = 1;
			dev_queue_xmit(skb);

			write_lock(&bond->ptrlock);
			bond->current_slave = slave->next;
			write_unlock(&bond->ptrlock);

			read_unlock_irqrestore(&bond->lock, flags);
			return 0;
		}
	} while ((slave = slave->next) != start_at);

	/* no suitable interface, frame not sent */
	dev_kfree_skb(skb);
	read_unlock_irqrestore(&bond->lock, flags);
	return 0;
}


/* in active-backup mode, we know that bond->current_slave is always valid if
   the bond has a usable interface.
*/
static int bond_xmit_activebackup(struct sk_buff *skb, struct net_device *dev)
{
	struct bonding *bond = (struct bonding *) dev->priv;
	unsigned long flags;
	int ret;

	if (!IS_UP(dev)) { /* bond down */
		dev_kfree_skb(skb);
		return 0;
	}

	read_lock_irqsave(&bond->lock, flags);

	read_lock(&bond->ptrlock);
	if (bond->current_slave != NULL) { /* one usable interface */
		skb->dev = bond->current_slave->dev;
		read_unlock(&bond->ptrlock);
		skb->priority = 1;
		ret = dev_queue_xmit(skb);
		read_unlock_irqrestore(&bond->lock, flags);
		return 0;
	}
	else {
		read_unlock(&bond->ptrlock);
	}

	/* no suitable interface, frame not sent */
#ifdef BONDING_DEBUG
	printk(KERN_INFO "There was no suitable interface, so we don't transmit\n");
#endif
	dev_kfree_skb(skb);
	read_unlock_irqrestore(&bond->lock, flags);
	return 0;
}

static struct net_device_stats *bond_get_stats(struct net_device *dev)
{
	bonding_t *bond = dev->priv;
	struct net_device_stats *stats = bond->stats, *sstats;
	slave_t *slave;
	unsigned long flags;

	memset(bond->stats, 0, sizeof(struct net_device_stats));

	read_lock_irqsave(&bond->lock, flags);

	for (slave = bond->prev; slave != (slave_t *)bond; slave = slave->prev) {
		sstats = slave->dev->get_stats(slave->dev);
 
		stats->rx_packets += sstats->rx_packets;
		stats->rx_bytes += sstats->rx_bytes;
		stats->rx_errors += sstats->rx_errors;
		stats->rx_dropped += sstats->rx_dropped;

		stats->tx_packets += sstats->tx_packets;
		stats->tx_bytes += sstats->tx_bytes;
		stats->tx_errors += sstats->tx_errors;
		stats->tx_dropped += sstats->tx_dropped;

		stats->multicast += sstats->multicast;
		stats->collisions += sstats->collisions;

		stats->rx_length_errors += sstats->rx_length_errors;
		stats->rx_over_errors += sstats->rx_over_errors;
		stats->rx_crc_errors += sstats->rx_crc_errors;
		stats->rx_frame_errors += sstats->rx_frame_errors;
		stats->rx_fifo_errors += sstats->rx_fifo_errors;	
		stats->rx_missed_errors += sstats->rx_missed_errors;
	
		stats->tx_aborted_errors += sstats->tx_aborted_errors;
		stats->tx_carrier_errors += sstats->tx_carrier_errors;
		stats->tx_fifo_errors += sstats->tx_fifo_errors;
		stats->tx_heartbeat_errors += sstats->tx_heartbeat_errors;
		stats->tx_window_errors += sstats->tx_window_errors;

	}

	read_unlock_irqrestore(&bond->lock, flags);
	return stats;
}

static int bond_get_info(char *buf, char **start, off_t offset, int length)
{
	bonding_t *bond = NULL;
	int len = 0;
	off_t begin = 0;
	int size;
	u16 link;
	slave_t *slave = NULL;

	if (this_bond == NULL) {
		return 0;
	}

	/*
	 * This function locks the mutex, so we can't lock it until afterwards
	 */
	link = bond_check_mii_link(this_bond);

	bond = (struct bonding *)this_bond->priv;
	read_lock(&bond->ptrlock);

	size = sprintf(buf + len, "Bonding Mode: ");
	len += size;
	size = sprintf(buf + len, "%s\n", mode ? "active-backup" : "load balancing");
	len += size;

	if (mode == BOND_MODE_ACTIVEBACKUP) {
		if (bond->current_slave != NULL) {
 			size = sprintf(buf + len, "Currently Active Slave: %s\n", bond->current_slave->dev->name);
			len += size;
		}
	}

	size = sprintf(buf + len, "MII Status: ");
	len += size;
 	size = sprintf(buf + len, link == MII_LINK_READY ? "up\n" : "down\n");
	len += size;
	size = sprintf(buf + len, "MII Polling Interval (ms): %d\n", miimon);
	len += size;
	size = sprintf(buf + len, "Up Delay (ms): %d\n", updelay);
	len += size;
	size = sprintf(buf + len, "Down Delay (ms): %d\n", downdelay);
	len += size;

	for (slave = bond->prev; 
		slave != (slave_t *)bond; 
		slave = slave->prev) {
		size = sprintf(buf + len, "\nSlave Interface: %s\n", slave->dev->name);
		len += size;

		size = sprintf(buf + len, "MII Status: ");
		len += size;

 		size = sprintf(buf + len, slave->link == BOND_LINK_UP ? "up\n" : "down\n");
		len += size;
 		size = sprintf(buf + len, "Link Failure Count: %d\n", slave->link_failure_count);
		len += size;
	}

	/*
	 * Figure out the calcs for the /proc/net interface
	 */
	*start = buf + (offset - begin);
	len -= (offset - begin);
	if (len > length) {
		len = length;
	}
	if (len < 0) {
		len = 0;
	}

	read_unlock(&bond->ptrlock);
	return len;
}

static int bond_event(struct notifier_block *this, unsigned long event, 
                      void *ptr)
{
	struct net_device *slave = ptr;

	if (this_bond == NULL 
	    || this_bond == slave 
	    || this_bond != slave->master)
		return NOTIFY_DONE;

	switch (event) {
	case NETDEV_UNREGISTER:
		bond_release(this_bond, slave);
		break;
	case NETDEV_CHANGEADDR:
		break;
	}
	return NOTIFY_DONE;
}

static struct notifier_block bond_netdev_notifier={
	bond_event,
	NULL,
	0
};

static int __init bond_init(struct net_device *dev)
{
	bonding_t *bond;

	bond = kmalloc(sizeof(struct bonding), GFP_KERNEL);
	if (bond == NULL) {
		return -ENOMEM;
	}
	memset(bond, 0, sizeof(struct bonding));

	/* initialize rwlocks */
	rwlock_init(&bond->lock);
	rwlock_init(&bond->ptrlock);
	
	bond->stats = kmalloc(sizeof(struct net_device_stats), GFP_KERNEL);
	if (bond->stats == NULL) {
		kfree(bond);
		return -ENOMEM;
	}
	memset(bond->stats, 0, sizeof(struct net_device_stats));

	bond->next = bond->prev = (slave_t *)bond;
	bond->current_slave = NULL;
	dev->priv = bond;

	/* Initialize the device structure. */
	if (mode == BOND_MODE_ACTIVEBACKUP) {
		dev->hard_start_xmit = bond_xmit_activebackup;
	}
	else {
		dev->hard_start_xmit = bond_xmit_roundrobin;
	}
	dev->get_stats	= bond_get_stats;
	dev->open = bond_open;
	dev->stop = bond_close;
	dev->set_multicast_list = set_multicast_list;
	dev->do_ioctl = bond_ioctl;

	/* Fill in the fields of the device structure with ethernet-generic 
	   values. */

	ether_setup(dev);

	dev->tx_queue_len = 0;
	dev->flags |= IFF_MASTER|IFF_MULTICAST;
#ifdef CONFIG_NET_FASTROUTE
	dev->accept_fastpath = bond_accept_fastpath;
#endif

	printk(KERN_INFO "%s registered with", dev->name);
	if (miimon > 0) {
		printk(" MII link monitoring set to %d ms", miimon);
		updelay /= miimon;
		downdelay /= miimon;
	}
	else
		printk("out MII link monitoring");

	printk(", in %s mode.\n",mode?"active-backup":"bonding");

	this_bond = dev;

#ifdef CONFIG_PROC_FS
	bond->bond_proc_dir = proc_mkdir(dev->name, proc_net);
	if (bond->bond_proc_dir == NULL) {
		printk(KERN_ERR "%s: Cannot init /proc/net/%s/\n", dev->name, dev->name);
		kfree(bond);
		return -ENOMEM;
	}
	bond->bond_proc_info_file = 
		create_proc_info_entry("info", 0, bond->bond_proc_dir, bond_get_info);
	if (bond->bond_proc_info_file == NULL) {
		printk(KERN_ERR "%s: Cannot init /proc/net/%s/info\n", dev->name, dev->name);
		remove_proc_entry(dev->name, proc_net);
		kfree(bond);
		return -ENOMEM;
	}
#endif /* CONFIG_PROC_FS */

	register_netdevice_notifier(&bond_netdev_notifier);

	return 0;
}

/*
static int __init bond_probe(struct net_device *dev)
{
	bond_init(dev);
	return 0;
}
 */

static struct net_device dev_bond = {
		"", /*bond_name,*/ 	/* Needs to be writeable */
		0, 0, 0, 0,
	 	0x0, 0,
	 	0, 0, 0, NULL, bond_init 
};

static int __init bonding_init(void)
{
	int err;

	SET_MODULE_OWNER(&dev_bond);
	
	/* Find a name for this unit */
	err = dev_alloc_name(&dev_bond,"bond%d");

	if (err < 0)
		return err;

	if (register_netdev(&dev_bond) != 0)
		return -EIO;

	return 0;
}

static void __exit bonding_exit(void)
{
	struct bonding *bond = (struct bonding *) dev_bond.priv;

	unregister_netdevice_notifier(&bond_netdev_notifier);
		 
#ifdef CONFIG_PROC_FS
	remove_proc_entry("info", bond->bond_proc_dir);
	remove_proc_entry(dev_bond.name, proc_net);
#endif
	unregister_netdev(&dev_bond);

	kfree(bond->stats);
	kfree(dev_bond.priv);

	dev_bond.priv = NULL;
}

module_init(bonding_init);
module_exit(bonding_exit);

/*
 * Local variables:
 *  c-indent-level: 8
 *  c-basic-offset: 8
 *  tab-width: 8
 * End:
 */
