#ifndef _LINUX_RWSEM_XCHGADD_H
#define _LINUX_RWSEM_XCHGADD_H

struct rw_semaphore
{
	long count;
	spinlock_t lock;
#define RWSEM_READ_BIAS 1
#define RWSEM_WRITE_BIAS (~(~0UL >> (BITS_PER_LONG>>1)))
	struct list_head wait;
#if RWSEM_DEBUG
	long __magic;
#endif
};

#if RWSEM_DEBUG
#define __SEM_DEBUG_INIT(name) \
	, (int)&(name).__magic
#define RWSEM_MAGIC(x)							\
	do {								\
		if ((x) != (long)&(x)) {				\
			printk("rwsem bad magic %lx (should be %lx), ",	\
				(long)x, (long)&(x));			\
			BUG();						\
		}							\
	} while (0)
#else
#define __SEM_DEBUG_INIT(name)
#define CHECK_MAGIC(x)
#endif

#define __RWSEM_INITIALIZER(name, count)	\
{						\
	(count),				\
	SPIN_LOCK_UNLOCKED,			\
	LIST_HEAD_INIT((name).wait)		\
	__SEM_DEBUG_INIT(name)			\
}
#define RWSEM_INITIALIZER(name) __RWSEM_INITIALIZER(name, 0)

#define __DECLARE_RWSEM(name, count) \
	struct rw_semaphore name = __RWSEM_INITIALIZER(name, count)
#define DECLARE_RWSEM(name) __DECLARE_RWSEM(name, 0)
#define DECLARE_RWSEM_READ_LOCKED(name) __DECLARE_RWSEM(name, RWSEM_READ_BIAS)
#define DECLARE_RWSEM_WRITE_LOCKED(name) __DECLARE_RWSEM(name, RWSEM_WRITE_BIAS+RWSEM_READ_BIAS)

#define RWSEM_READ_BLOCKING_BIAS (RWSEM_WRITE_BIAS-RWSEM_READ_BIAS)
#define RWSEM_WRITE_BLOCKING_BIAS (-RWSEM_READ_BIAS)

#define RWSEM_READ_MASK (~RWSEM_WRITE_BIAS)
#define RWSEM_WRITE_MASK (RWSEM_WRITE_BIAS)

/*
 * We return the semaphore itself from the C functions so we can pass it
 * in %eax via regparm and we don't need to declare %eax clobbered by C.
 * This is mostly for x86 but maybe other archs can make a use of it too.
 * Idea is from David Howells <dhowells@redhat.com>.
 */
extern struct rw_semaphore * FASTCALL(rwsem_down_failed(struct rw_semaphore *, long));
extern struct rw_semaphore * FASTCALL(rwsem_wake(struct rw_semaphore *));

static inline void init_rwsem(struct rw_semaphore *sem)
{
	sem->count = 0;
	spin_lock_init(&sem->lock);
	INIT_LIST_HEAD(&sem->wait);
#if RWSEM_DEBUG
	sem->__magic = (long)&sem->__magic;
#endif
}

#include <asm/rwsem_xchgadd.h>

static inline void down_read(struct rw_semaphore *sem)
{
	CHECK_MAGIC(sem->__magic);

	__down_read(sem);
}

static inline void down_write(struct rw_semaphore *sem)
{
	CHECK_MAGIC(sem->__magic);

	__down_write(sem);
}

static inline void up_read(struct rw_semaphore *sem)
{
	CHECK_MAGIC(sem->__magic);

	__up_read(sem);
}

static inline void up_write(struct rw_semaphore *sem)
{
	CHECK_MAGIC(sem->__magic);

	__up_write(sem);
}

#endif /* _LINUX_RWSEM_XCHGADD_H */
