/*
 *  rw_semaphores generic spinlock version
 *  Copyright (C) 2001 Andrea Arcangeli <andrea@suse.de> SuSE
 */

#include <linux/sched.h>
#include <linux/module.h>
#include <asm/semaphore.h>

struct rwsem_wait_queue {
	unsigned long retire;
	struct task_struct * task;
	struct list_head task_list;
};

static void FASTCALL(rwsem_down_failed(struct rw_semaphore *, long));
static void rwsem_down_failed(struct rw_semaphore *sem, long retire)
{
	struct task_struct *tsk = current;
	struct rwsem_wait_queue wait;

	sem->count += retire;
	wait.retire = retire;
	wait.task = tsk;
	INIT_LIST_HEAD(&wait.task_list);
	list_add(&wait.task_list, &sem->wait);

	do {
		__set_task_state(tsk, TASK_UNINTERRUPTIBLE);
		spin_unlock(&sem->lock);
		schedule();
		spin_lock(&sem->lock);
	} while(wait.task_list.next);
}

static void FASTCALL(rwsem_wake(struct rw_semaphore *));
static void rwsem_wake(struct rw_semaphore *sem)
{
	struct list_head * entry, * head = &sem->wait;
	int last = 0;

	while ((entry = head->prev) != head) {
		struct rwsem_wait_queue * wait;

		wait = list_entry(entry, struct rwsem_wait_queue, task_list);

		if (wait->retire == RWSEM_WRITE_BLOCKING_BIAS) {
			if (sem->count & RWSEM_READ_MASK)
				break;
			last = 1;
		}

		/* convert write lock into read lock when read become active */
		sem->count -= wait->retire;
		list_del(entry);
		entry->next = NULL;
		wake_up_process(wait->task);
			
		if (last)
			break;
	}
}

void init_rwsem(struct rw_semaphore *sem)
{
	spin_lock_init(&sem->lock);
	sem->count = 0;
	INIT_LIST_HEAD(&sem->wait);
#if RWSEM_DEBUG
	sem->__magic = (long)&sem->__magic;
#endif
}

void down_read(struct rw_semaphore *sem)
{
	CHECK_MAGIC(sem->__magic);

	spin_lock(&sem->lock);
	sem->count += RWSEM_READ_BIAS;
	if (__builtin_expect(sem->count, 0) < 0)
		rwsem_down_failed(sem, RWSEM_READ_BLOCKING_BIAS);
	spin_unlock(&sem->lock);
}

void down_write(struct rw_semaphore *sem)
{
	long count;
	CHECK_MAGIC(sem->__magic);

	spin_lock(&sem->lock);
	count = sem->count;
	sem->count += RWSEM_WRITE_BIAS;
	if (__builtin_expect(count, 0))
		rwsem_down_failed(sem, RWSEM_WRITE_BLOCKING_BIAS);
	spin_unlock(&sem->lock);
}

void up_read(struct rw_semaphore *sem)
{
	CHECK_MAGIC(sem->__magic);

	spin_lock(&sem->lock);
	sem->count -= RWSEM_READ_BIAS;
	if (__builtin_expect(sem->count < 0 && !(sem->count & RWSEM_READ_MASK), 0))
		rwsem_wake(sem);
	spin_unlock(&sem->lock);
}

void up_write(struct rw_semaphore *sem)
{
	CHECK_MAGIC(sem->__magic);

	spin_lock(&sem->lock);
	sem->count -= RWSEM_WRITE_BIAS;
	if (__builtin_expect(sem->count, 0))
		rwsem_wake(sem);
	spin_unlock(&sem->lock);
}

EXPORT_SYMBOL(init_rwsem);
EXPORT_SYMBOL(down_read);
EXPORT_SYMBOL(down_write);
EXPORT_SYMBOL(up_read);
EXPORT_SYMBOL(up_write);
